// zlib License. Warzone2100. wz2100.net - official site.

#include "CreateDxtFile.h"

static void wzpng_read_data(png_structp ctx, png_bytep area, png_size_t size)
{
	PHYSFS_file *fileHandle = (PHYSFS_file *)png_get_io_ptr(ctx);

	PHYSFS_read(fileHandle, area, 1, size);
}

static inline void PNGReadCleanup(png_infop *info_ptr, png_structp *png_ptr, PHYSFS_file *fileHandle)
{
	if (*info_ptr != NULL)
	{
		png_destroy_info_struct(*png_ptr, info_ptr);
	}
	if (*png_ptr != NULL)
	{
		png_destroy_read_struct(png_ptr, NULL, NULL);
	}
	if (fileHandle != NULL)
	{
		PHYSFS_close(fileHandle);
	}
}

bool iV_loadImage_PNG(const char *fileName, int &outWidth, int &outHeight, int &depth, bool &outHasAlpha, unsigned char **outData)
{

	unsigned char PNGheader[8];
	PHYSFS_sint64 readSize;

	png_structp png_ptr = NULL;
	png_infop info_ptr = NULL;

	// Open file
	PHYSFS_file *fileHandle = PHYSFS_openRead(fileName);
	if (fileHandle == NULL)
	{
	//	debug(LOG_ERROR, "pie_PNGLoadFile: PHYSFS_openRead(%s) failed with error: %s\n", fileName, PHYSFS_getLastError());
		PNGReadCleanup(&info_ptr, &png_ptr, fileHandle);
		return false;
	}

	// Read PNG header from file
	readSize = PHYSFS_read(fileHandle, PNGheader, 1, 8);
	if (readSize < 8)
	{
	//	debug(LOG_ERROR, "pie_PNGLoadFile: PHYSFS_read(%s) failed with error: %s\n", fileName, PHYSFS_getLastError());
		PNGReadCleanup(&info_ptr, &png_ptr, fileHandle);
		return false;
	}

	// Verify the PNG header to be correct
	if (png_sig_cmp(PNGheader, 0, 8))
	{
	//	debug(LOG_3D, "pie_PNGLoadMem: Did not recognize PNG header in %s", fileName);
		PNGReadCleanup(&info_ptr, &png_ptr, fileHandle);
		return false;
	}

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (png_ptr == NULL)
	{
	//	debug(LOG_3D, "pie_PNGLoadMem: Unable to create png struct");
		PNGReadCleanup(&info_ptr, &png_ptr, fileHandle);
		return false;
	}

	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL)
	{
	//	debug(LOG_3D, "pie_PNGLoadMem: Unable to create png info struct");
		PNGReadCleanup(&info_ptr, &png_ptr, fileHandle);
		return false;
	}

	// Set libpng's failure jump position to the if branch,
	// setjmp evaluates to false so the else branch will be executed at first
	if (setjmp(png_jmpbuf(png_ptr)))
	{
	//	debug(LOG_3D, "pie_PNGLoadMem: Error decoding PNG data in %s", fileName);
		PNGReadCleanup(&info_ptr, &png_ptr, fileHandle);
		return false;
	}

	// Tell libpng how many byte we already read
	png_set_sig_bytes(png_ptr, 8);

	/* Set up the input control */
	png_set_read_fn(png_ptr, fileHandle, wzpng_read_data);

	// Most of the following transformations are seemingly not needed
	// Filler is, however, for an unknown reason required for tertilesc[23]

	/* tell libpng to strip 16 bit/color files down to 8 bits/color */
	png_set_strip_16(png_ptr);

	/* Extract multiple pixels with bit depths of 1, 2, and 4 from a single
	 * byte into separate bytes (useful for paletted and grayscale images).
	 */
// 	png_set_packing(png_ptr);

	/* More transformations to ensure we end up with 32bpp, 4 channel RGBA */
	png_set_gray_to_rgb(png_ptr);
	png_set_palette_to_rgb(png_ptr);
	png_set_tRNS_to_alpha(png_ptr);
	png_set_filler(png_ptr, 0xff, PNG_FILLER_AFTER);
//	png_set_gray_1_2_4_to_8(png_ptr);

	png_read_png(png_ptr, info_ptr, PNG_TRANSFORM_IDENTITY, NULL);

	outWidth = png_get_image_width(png_ptr, info_ptr);
	outHeight = png_get_image_height(png_ptr, info_ptr);
	depth = png_get_channels(png_ptr, info_ptr);
	*outData = (unsigned char *)malloc(outHeight * png_get_rowbytes(png_ptr, info_ptr));

	{
		unsigned int i = 0;
		png_bytepp row_pointers = png_get_rows(png_ptr, info_ptr);
		for (i = 0; i < png_get_image_height(png_ptr, info_ptr); i++)
		{
			memcpy(*outData + (png_get_rowbytes(png_ptr, info_ptr) * i), row_pointers[i], png_get_rowbytes(png_ptr, info_ptr));
			
		}
	}

	PNGReadCleanup(&info_ptr, &png_ptr, fileHandle);
	//ASSERT_OR_RETURN(false, image->depth > 3, "Unsupported image depth (%d) found.  We only support 3 (RGB) or 4 (ARGB)", image->depth);
	return true;
}

bool CreateDxtFile(std::string DIR, const char* OneFile, int format, bool create_mipmap, short setup_min_size, int &begin_identifier, bool AfterDeletePngFile, bool rename_extension)
{
	int kDxt1 = 1;
	int kDxt3 = 2;
	int kDxt5 = 4;
	bool hasAlpha=0;
	int width_png, height_png, depth=0;
	unsigned char *textureImage;
	bool success = iV_loadImage_PNG(OneFile, width_png, height_png, depth, hasAlpha, &textureImage);
	if (!success)
	{
        std::cerr << "error open PNG file" << std::endl;
        return 0;
    }
	if(AfterDeletePngFile)
	{
			string f = DIR + "/" + OneFile;
			remove(f.c_str());
	}

	std::cout << begin_identifier << "   "  << OneFile<<"\n";

	int wid = width_png;
	int level_before = 0;
	int dxt_type = format;


	while(1)
	{
		if(wid == 4) break;		
		wid=wid/2;
		level_before++;
	}
	if(create_mipmap == 0) level_before = 0;
	// std::cout << level_before << "\n";

    std::string ptrstr = OneFile;
	std::string juststr;

	if(rename_extension) 
	{
		int pos = ptrstr.find(".png",0);
	    ptrstr.replace(pos, 4, ".dxt");
		juststr = DIR + "/" + ptrstr;
	}
	else
	{
		juststr = DIR + "/" + OneFile + ".dxt";
	}


	std::ofstream out(juststr.c_str(), std::ios::out | std::ios::binary);

	char filetype[4] = {'D','X','T',' '};
	out.write((char *) &filetype, 4);

	unsigned char type = dxt_type;
	out.write((char *) &type, 1);

	unsigned char level = level_before;
	out.write((char *) &level, 1);

	short identifierOfTexture = begin_identifier;   
	unsigned char identifier[2] = {(unsigned char)(identifierOfTexture&0xff),(unsigned char)(identifierOfTexture>>8)};
	out.write((char *) &identifier, 2);

	unsigned char dxt_width[2] = {(unsigned char)(width_png&0xff),(unsigned char)(width_png>>8)};	
	out.write((char *) &dxt_width, 2);

	unsigned char dxt_height[2] = {(unsigned char)(height_png&0xff),(unsigned char)(height_png>>8)};	
	out.write((char *) &dxt_height, 2);

	//unsigned char dxt_size[4] = {(unsigned char)(__dxt_size&0xff),(unsigned char)(__dxt_size>>8),
	//	                           (unsigned char)(__dxt_size>>16),(unsigned char)(__dxt_size>>24)};	


	ZoomOutTwice k ;
	k.width =  width_png;      // only 32x32, 64x64, 128x128, 256x256, 512x512, 1024x1024, 2048x2048, 4096x4096
    k.height = height_png;
	k.format = GL_RGBA;
	k.DataIn = textureImage;
	k.SizeEnd = (char)setup_min_size;              


    while(1)
    {
	   if(k.ReduceImageByHalf() != 1) break;

	   int __dxt_size;

	   if(dxt_type == kDxt1) 
	   {
		   __dxt_size = ((int)k.width * (int)k.height)/2;
	   }
	   else
	   {
		   __dxt_size = ((int)k.width * (int)k.height);
	   }
	   unsigned char* dxtarray = (unsigned char*)malloc((__dxt_size) * sizeof(unsigned char));

	   CompressImage(k.DataOut,(int)k.width,(int)k.height,dxtarray,dxt_type);

	   out.write((const char*)dxtarray, __dxt_size);
	   dxtarray = nullptr;
       if(create_mipmap == 0) break;
	}  

	out.close ();

	textureImage = nullptr;
	begin_identifier++;
	return true;
}

bool CreateDxtfilesFolder(std::string DIR, int format, bool create_mipmap, short setup_min_size, int &begin_identifier, bool AfterDeletePngFile, bool rename_extension)
{
	tinydir_dir dir;
	int kDxt1 = 1;
	int kDxt3 = 2;
	int kDxt5 = 4;
    wstring str = std::wstring(DIR.begin(),DIR.end());

    tinydir_open(&dir, str.c_str());

	unsigned char *textureImage;

    while(dir.has_next)
    {
   	    tinydir_file file;
	    tinydir_readfile(&dir,&file);
	    _bstr_t x(file.name);

	    if(file.is_dir)  // this is name of folder
	    {
	    //	std::cout << x;     /
	    //	std::cout << "/";
    	//  std::cout << "\n";
     	}
	    else
	    {
		    bool hasAlpha=0;
		    int width_png, height_png, depth=0;

	        string str =  (string)x;	bool success = iV_loadImage_PNG(str.c_str(), width_png, height_png, depth, hasAlpha, &textureImage);

	    	if (!success)
		    {
                std::cerr << "error open PNG file  - " << str.c_str() << std::endl;
                goto m1;
            }
			if(AfterDeletePngFile)
	        {
			   string f = DIR + "/" + str;
			   remove(f.c_str());
	        }

			std::cout << begin_identifier << "   " << str.c_str() <<"\n";

		    int wid = width_png;
	    	int level_before = 0;
		    int dxt_type = format;


		    while(1)
		    {
		    	if(wid == 4) break;		
		    	wid=wid/2;

		    	level_before++;
		    }
		    if(create_mipmap == 0) level_before = 0;
		    // std::cout << level_before << "\n";


	        std::string juststr;
	        if(rename_extension) 
	        {
	        	int pos = str.find(".png",0);
	            str.replace(pos, 4, ".dxt");
	        	juststr = DIR + "/" + str;
	        }
	        else
	        {
	         	juststr = DIR + "/" + str + ".dxt";
          	}

		    std::ofstream out(juststr.c_str(), std::ios::out | std::ios::binary);

		    char filetype[4] = {'D','X','T',' '};
		    out.write((char *) &filetype, 4);

		    unsigned char type = dxt_type;
		    out.write((char *) &type, 1);

		    unsigned char level = level_before;
		    out.write((char *) &level, 1);

		    short identifierOfTexture = begin_identifier;   
		    unsigned char identifier[2] = {(unsigned char)(identifierOfTexture&0xff),(unsigned char)(identifierOfTexture>>8)};
		    out.write((char *) &identifier, 2);

		    unsigned char dxt_width[2] = {(unsigned char)(width_png&0xff),(unsigned char)(width_png>>8)};	
		    out.write((char *) &dxt_width, 2);

		    unsigned char dxt_height[2] = {(unsigned char)(height_png&0xff),(unsigned char)(height_png>>8)};	
		    out.write((char *) &dxt_height, 2);

		    //unsigned char dxt_size[4] = {(unsigned char)(__dxt_size&0xff),(unsigned char)(__dxt_size>>8),
		    //	                           (unsigned char)(__dxt_size>>16),(unsigned char)(__dxt_size>>24)};	

	        ZoomOutTwice k ;
	        k.width =  width_png;      // only 32x32, 64x64, 128x128, 256x256, 512x512, 1024x1024, 2048x2048, 4096x4096
            k.height = height_png;
	        k.format = GL_RGBA;
	        k.DataIn = textureImage;
	        k.SizeEnd = (char)setup_min_size;              


    	    while(1)
    	    {
	        	if(k.ReduceImageByHalf() != 1) break;

	        	int __dxt_size;

		        if(dxt_type == kDxt1) 
		        {
		        	__dxt_size = ((int)k.width * (int)k.height)/2;
		        }
	    	    else
		        {
		        	__dxt_size = ((int)k.width * (int)k.height);
		        }
		        unsigned char* dxtarray = (unsigned char*)malloc((__dxt_size) * sizeof(unsigned char));

		        CompressImage(k.DataOut,(int)k.width,(int)k.height,dxtarray,dxt_type);

		        out.write((const char*)dxtarray, __dxt_size);
				dxtarray = nullptr;
                if(create_mipmap == 0) break;
	       }  

	     out.close ();
		 textureImage = nullptr;
		 begin_identifier++;

	    }
m1:
	  tinydir_next(&dir);  
    }

   tinydir_close(&dir);
   return true;
}
